from pathlib import Path
from typing import Any, List, Set, Union

def chdir(directory: Union[str, Path]) -> Any: ...
def remove_util(p: Union[str, Path]) -> None: ...

class PNGProvider:
    bitmaps_dir: Path
    def __init__(self, bitmaps_dir: Union[str, Path]) -> None: ...
    def get(self, key: str) -> Union[List[Path], Path]: ...

def add_missing_xcursors(directory: Path, data: List[Set[str]]=..., rename: bool=..., force: bool=...) -> None: ...
def timer(func: Any): ...
def debug(func: Any): ...
